 drivers/char/random.c | 55 ++++++++++++++++++++++++++++++++++++++++++++++++---
 1 file changed, 52 insertions(+), 3 deletions(-)

diff --git a/drivers/char/random.c b/drivers/char/random.c
index 5d5ea4ce1442..0d7dc86e0f60 100644
--- a/drivers/char/random.c
+++ b/drivers/char/random.c
@@ -2118,6 +2118,55 @@ const struct file_operations urandom_fops = {
 	.llseek = noop_llseek,
 };
 
+/*
+ * We refuse to wait very long for a blocking getrandom().
+ *
+ * The crng may not be ready during boot, but if you ask for
+ * blocking random numbers very early, there is no guarantee
+ * that you'll ever get any timely entropy.
+ *
+ * If you are sure you need entropy and that you can generate
+ * it, you need to ask for non-blocking random state, and then
+ * if that fails you must actively _do_something_ that causes
+ * enough system activity, perhaps asking the user to type
+ * something on the keyboard.
+ *
+ * Just asking for blocking random numbers is completely and
+ * fundamentally wrong, and the kernel will not play that game.
+ *
+ * We will block for at most 15 seconds at a time, and if called
+ * sequentially will decrease the blocking amount so that we'll
+ * block for at most 30s total - and if people continue to ask
+ * for blocking, at that point we'll just return whatever random
+ * state we have acquired.
+ *
+ * This will also complain loudly if the timeout happens, to let
+ * the distribution or system admin know about the problem.
+ *
+ * The process that gets the -EAGAIN will hopefully also log the
+ * error, to raise awareness that there may be use of random
+ * numbers without sufficient entropy.
+ */
+static int getrandom_wait(void)
+{
+	static unsigned int getrandom_timeout = 15*HZ;
+	unsigned int timeout = READ_ONCE(getrandom_timeout);
+	int ret;
+
+	/* "At some point you just have to give up on broken boot scripts" */
+	if (!timeout)
+		return 0;
+
+	ret = wait_event_interruptible_timeout(crng_init_wait, crng_ready(), timeout);
+	/* Success (>0) or interrupted (<0)? */
+	if (likely(ret))
+		return ret > 0 ? 0 : ret;
+
+	WRITE_ONCE(getrandom_timeout, timeout >> 1);
+	WARN_ONCE(1, "getrandom() timed out with no entropy");
+	return -EAGAIN;
+}
+
 SYSCALL_DEFINE3(getrandom, char __user *, buf, size_t, count,
 		unsigned int, flags)
 {
@@ -2132,11 +2181,11 @@ SYSCALL_DEFINE3(getrandom, char __user *, buf, size_t, count,
 	if (flags & GRND_RANDOM)
 		return _random_read(flags & GRND_NONBLOCK, buf, count);
 
-	if (!crng_ready()) {
+	if (unlikely(!crng_ready())) {
 		if (flags & GRND_NONBLOCK)
 			return -EAGAIN;
-		ret = wait_for_random_bytes();
-		if (unlikely(ret))
+		ret = getrandom_wait();
+		if (ret)
 			return ret;
 	}
 	return urandom_read(NULL, buf, count, NULL);
